"""Integration tests for metadata gathering using real resource files

Tests in this file should use run_job()
"""

import json
import os
import unittest
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict
from unittest.mock import MagicMock, patch

from aind_data_schema.core.acquisition import Acquisition
from aind_data_schema.core.data_description import DataDescription
from aind_data_schema.core.instrument import Instrument
from aind_data_schema.core.metadata import Metadata
from aind_data_schema.core.procedures import Procedures
from aind_data_schema.core.processing import Processing
from aind_data_schema.core.quality_control import QualityControl
from aind_data_schema.core.subject import Subject
from aind_data_schema_models.modalities import Modality

from aind_metadata_mapper.gather_metadata import GatherMetadataJob
from aind_metadata_mapper.models import DataDescriptionSettings, InstrumentSettings, JobSettings

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
METADATA_SERVICE_DIR = TEST_DIR / "resources" / "metadata_service"
V2_METADATA_DIR = TEST_DIR / "resources" / "v2_metadata"


class TestIntegrationMetadata(unittest.TestCase):
    """Integration tests using real metadata resource files"""

    @patch("os.makedirs")
    def setUp(self, mock_makedirs):
        """Set up test fixtures"""
        self.test_settings = JobSettings(
            metadata_dir="/test/metadata",
            output_dir="/test/output",
            subject_id="804670",
            data_description_settings=DataDescriptionSettings(
                project_name="Visual Behavior",
                modalities=[Modality.BEHAVIOR, Modality.ECEPHYS],
            ),
            metadata_service_url="http://test-service.com",
            acquisition_start_time=datetime(2025, 9, 17, 10, 26, 0, tzinfo=timezone.utc),
        )
        self.job = GatherMetadataJob(settings=self.test_settings)

    def _load_resource_file(self, resource_dir: Path, filename: str) -> Dict[str, Any]:  # noqa: E501
        """Helper method to load JSON resource files"""
        file_path = resource_dir / filename
        with open(file_path, "r") as f:
            return json.load(f)

    def test_metadata_service_responses_structure(self):
        """Test that metadata service response files have expected structure"""
        # Test funding response structure
        funding_data = self._load_resource_file(METADATA_SERVICE_DIR, "funding_response.json")
        self.assertIsInstance(funding_data, list)
        self.assertGreater(len(funding_data), 0)

        first_funding = funding_data[0]
        self.assertEqual(first_funding["object_type"], "Funding")
        self.assertIn("funder", first_funding)

        # Test subject response structure
        subject_data = self._load_resource_file(METADATA_SERVICE_DIR, "subject_response.json")
        self.assertEqual(subject_data["object_type"], "Subject")
        self.assertEqual(subject_data["subject_id"], "804670")
        self.assertIn("subject_details", subject_data)

        # Test procedures response structure
        procedures_data = self._load_resource_file(METADATA_SERVICE_DIR, "procedures_response.json")
        self.assertEqual(procedures_data["object_type"], "Procedures")
        self.assertEqual(procedures_data["subject_id"], "804670")
        self.assertIn("subject_procedures", procedures_data)

    def test_v2_metadata_files_structure(self):
        """Test that v2 metadata files have expected structure"""
        # Test data description
        data_desc = self._load_resource_file(V2_METADATA_DIR, "data_description.json")
        self.assertEqual(data_desc["object_type"], "Data description")
        self.assertEqual(data_desc["subject_id"], "804670")
        self.assertIn("funding_source", data_desc)
        self.assertIn("investigators", data_desc)

        # Test subject
        subject = self._load_resource_file(V2_METADATA_DIR, "subject.json")
        self.assertEqual(subject["object_type"], "Subject")
        self.assertEqual(subject["subject_id"], "804670")

        # Test acquisition
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        self.assertEqual(acquisition["object_type"], "Acquisition")
        self.assertEqual(acquisition["subject_id"], "804670")

        # Test procedures
        procedures = self._load_resource_file(V2_METADATA_DIR, "procedures.json")
        self.assertEqual(procedures["object_type"], "Procedures")
        self.assertEqual(procedures["subject_id"], "804670")

        # Test instrument
        instrument = self._load_resource_file(V2_METADATA_DIR, "instrument.json")
        self.assertEqual(instrument["object_type"], "Instrument")

        # Test processing
        processing = self._load_resource_file(V2_METADATA_DIR, "processing.json")
        self.assertEqual(processing["object_type"], "Processing")

        # Test quality control
        quality_control = self._load_resource_file(V2_METADATA_DIR, "quality_control.json")
        self.assertEqual(quality_control["object_type"], "Quality control")

    def test_validate_data_description_schema(self):
        """Test DataDescription validation"""
        data_desc_data = self._load_resource_file(V2_METADATA_DIR, "data_description.json")
        data_desc = DataDescription.model_validate(data_desc_data)
        self.assertIsInstance(data_desc, DataDescription)
        self.assertEqual(data_desc.subject_id, "804670")

    def test_validate_subject_schema(self):
        """Test Subject validation"""
        subject_data = self._load_resource_file(V2_METADATA_DIR, "subject.json")
        subject = Subject.model_validate(subject_data)
        self.assertIsInstance(subject, Subject)
        self.assertEqual(subject.subject_id, "804670")

    def test_validate_acquisition_schema(self):
        """Test Acquisition validation"""
        acquisition_data = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        acquisition = Acquisition.model_validate(acquisition_data)
        self.assertIsInstance(acquisition, Acquisition)
        self.assertEqual(acquisition.subject_id, "804670")

    def test_validate_procedures_schema(self):
        """Test Procedures validation"""
        procedures_data = self._load_resource_file(V2_METADATA_DIR, "procedures.json")
        procedures = Procedures.model_validate(procedures_data)
        self.assertIsInstance(procedures, Procedures)
        self.assertEqual(procedures.subject_id, "804670")

    def test_validate_instrument_schema(self):
        """Test Instrument validation"""
        instrument_data = self._load_resource_file(V2_METADATA_DIR, "instrument.json")
        instrument = Instrument.model_validate(instrument_data)
        self.assertIsInstance(instrument, Instrument)

    def test_validate_processing_schema(self):
        """Test Processing validation"""
        processing_data = self._load_resource_file(V2_METADATA_DIR, "processing.json")
        processing = Processing.model_validate(processing_data)
        self.assertIsInstance(processing, Processing)

    def test_validate_quality_control_schema(self):
        """Test QualityControl validation"""
        qc_data = self._load_resource_file(V2_METADATA_DIR, "quality_control.json")
        quality_control = QualityControl.model_validate(qc_data)
        self.assertIsInstance(quality_control, QualityControl)

    def test_validate_complete_metadata_object(self):
        """Test that v2 metadata files create a valid Metadata object"""
        # Load all metadata files
        data_description = self._load_resource_file(V2_METADATA_DIR, "data_description.json")
        subject = self._load_resource_file(V2_METADATA_DIR, "subject.json")
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        procedures = self._load_resource_file(V2_METADATA_DIR, "procedures.json")
        instrument = self._load_resource_file(V2_METADATA_DIR, "instrument.json")
        processing = self._load_resource_file(V2_METADATA_DIR, "processing.json")
        quality_control = self._load_resource_file(V2_METADATA_DIR, "quality_control.json")

        # Create core metadata dictionary
        core_metadata = {
            "data_description": data_description,
            "subject": subject,
            "acquisition": acquisition,
            "procedures": procedures,
            "instrument": instrument,
            "processing": processing,
            "quality_control": quality_control,
        }

        # Test validation through the job's validate_and_create_metadata method
        metadata = self.job.validate_and_create_metadata(core_metadata)
        self.assertIsInstance(metadata, (Metadata, dict))

        # If it's a Metadata object, check its properties
        if isinstance(metadata, Metadata):
            self.assertEqual(metadata.name, data_description["name"])
            self.assertIsNotNone(metadata.subject)
            self.assertIsNotNone(metadata.data_description)
            self.assertIsNotNone(metadata.acquisition)
            self.assertIsNotNone(metadata.procedures)
            self.assertIsNotNone(metadata.instrument)
            self.assertIsNotNone(metadata.processing)
            self.assertIsNotNone(metadata.quality_control)

    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob." + "_does_file_exist_in_user_defined_dir")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob." + "_get_file_from_user_defined_directory")
    def test_run_job_with_all_local_files(self, mock_get_file, mock_file_exists):
        """Test run_job method using all v2 metadata files as local files"""
        # Load all resource files
        v2_files = {
            "data_description.json": self._load_resource_file(V2_METADATA_DIR, "data_description.json"),
            "subject.json": self._load_resource_file(V2_METADATA_DIR, "subject.json"),
            "acquisition.json": self._load_resource_file(V2_METADATA_DIR, "acquisition.json"),
            "procedures.json": self._load_resource_file(V2_METADATA_DIR, "procedures.json"),
            "instrument.json": self._load_resource_file(V2_METADATA_DIR, "instrument.json"),
            "processing.json": self._load_resource_file(V2_METADATA_DIR, "processing.json"),
            "quality_control.json": self._load_resource_file(V2_METADATA_DIR, "quality_control.json"),
        }

        # Mock file existence - return True for files that exist locally
        def mock_exists(file_name):
            """Mock file existence check"""
            return file_name in v2_files or file_name == "data_description.json"

        mock_file_exists.side_effect = mock_exists

        # Mock file reading
        def mock_read_file(file_name):
            """Mock file reading"""
            if file_name in v2_files:
                return v2_files[file_name]
            elif file_name == "data_description.json":
                # Created by build_data_description, won't be called
                return v2_files["data_description.json"]
            else:
                raise FileNotFoundError(f"File {file_name} not found")

        mock_get_file.side_effect = mock_read_file

        # Mock file writing methods to avoid actual file writing
        with patch.object(self.job, "_write_json_file") as mock_write:
            # Run the job
            try:
                self.job.run_job()

                # Verify that files were "written" (mocked)
                written_files = [call[0][0] for call in mock_write.call_args_list]

                # Check that individual metadata files were written through _write_json_file
                expected_files = [
                    "data_description.json",
                    "subject.json",
                    "procedures.json",
                    "acquisition.json",
                    "instrument.json",
                    "processing.json",
                    "quality_control.json",
                ]

                for expected_file in expected_files:
                    self.assertIn(expected_file, written_files, f"Expected {expected_file} to be written")

            except Exception as e:
                self.fail(f"run_job with all local files failed: {e}")

    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob." + "_does_file_exist_in_user_defined_dir")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob." + "_get_file_from_user_defined_directory")
    def test_build_data_description_uses_acquisition_start_time_for_name(self, mock_get_file, mock_file_exists):
        """Test that build_data_description uses acquisition_start_time for DataDescription name formatting"""
        mock_file_exists.return_value = False
        acquisition_data = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")

        def mock_read_file(file_name):
            """Mock file reading"""
            if file_name == "acquisition.json":
                return acquisition_data
            else:
                raise FileNotFoundError(f"File {file_name} not found")

        mock_get_file.side_effect = mock_read_file

        funding_data = self._load_resource_file(METADATA_SERVICE_DIR, "funding_response.json")
        investigators_data = self._load_resource_file(METADATA_SERVICE_DIR, "investigators_response.json")

        with patch("requests.get") as mock_get:

            def mock_response_side_effect(url, *args, **kwargs):
                """Mock requests.get responses based on URL"""
                mock_response = MagicMock()
                mock_response.status_code = 200
                if "funding" in url:
                    mock_response.json.return_value = funding_data
                elif "investigators" in url:
                    mock_response.json.return_value = investigators_data
                return mock_response

            mock_get.side_effect = mock_response_side_effect

            acquisition_start_time = acquisition_data["acquisition_start_time"]
            subject_id = acquisition_data["subject_id"]
            result = self.job.build_data_description(
                acquisition_start_time=acquisition_start_time, subject_id=subject_id
            )

        self.assertIn("name", result)
        self.assertIn("creation_time", result)

        expected_name = "804670_2025-09-17_10-26-00"
        self.assertEqual(result["name"], expected_name)

        from datetime import datetime

        now = datetime.now()
        creation_time_str = result["creation_time"]
        self.assertIn(str(now.year), creation_time_str)
        self.assertIn(f"{now.month:02d}", creation_time_str)

    def test_validate_and_create_metadata_with_raise_if_invalid_false(self):
        """Test validate_and_create_metadata with raise_if_invalid=False (default) - prints errors but not raise"""
        # Create core metadata with invalid data that will cause ValidationError
        core_metadata = {
            "data_description": {
                "name": "test_dataset",
                "creation_time": "2023-01-01T12:00:00",
                "institution": {"name": "Allen Institute for Neural Dynamics"},
                "data_level": "raw",
                "modalities": ["Behavior videos"],
                "project_name": "Test Project",
                "funding_source": [{"funder": "Test Funder"}],
                "investigators": [{"name": "Test Investigator"}],
                "subject_id": "invalid_subject_id",  # This will cause validation issues
            },
            "subject": {
                "subject_id": "different_subject_id",  # Mismatch with data_description will cause validation error
                "sex": "Invalid Sex Value",  # Invalid enum value
                "date_of_birth": "invalid-date-format",  # Invalid date format
            },
            "acquisition": {
                "subject_id": "yet_another_subject_id",  # Another mismatch
                "acquisition_start_time": "invalid-datetime",  # Invalid datetime
            },
        }

        # Capture logging output to verify error messages are logged
        from unittest.mock import patch

        with patch("logging.warning") as mock_warning:
            result = self.job.validate_and_create_metadata(core_metadata)

            self.assertIsNotNone(result)

            mock_warning.assert_called()
            warning_calls = [call[0][0] for call in mock_warning.call_args_list]
            self.assertTrue(any("Validation Errors Found:" in call for call in warning_calls))

    def test_validate_and_create_metadata_with_raise_if_invalid_true(self):
        """Test validate_and_create_metadata with raise_if_invalid=True - should raise ValidationError"""
        with patch("os.makedirs"):
            strict_settings = JobSettings(
                metadata_dir="/test/metadata",
                output_dir="/test/output",
                subject_id="804670",
                data_description_settings=DataDescriptionSettings(
                    project_name="Visual Behavior",
                    modalities=[Modality.BEHAVIOR, Modality.ECEPHYS],
                ),
                metadata_service_url="http://test-service.com",
                raise_if_invalid=True,
                acquisition_start_time=datetime(2025, 9, 17, 10, 26, 0),
            )
            strict_job = GatherMetadataJob(settings=strict_settings)

        # Create core metadata with invalid data that will cause ValidationError
        core_metadata = {
            "data_description": {
                "name": "test_dataset",
                "creation_time": "2023-01-01T12:00:00",
                "institution": {"name": "Allen Institute for Neural Dynamics"},
                "data_level": "raw",
                "modalities": ["Behavior videos"],
                "project_name": "Test Project",
                "funding_source": [{"funder": "Test Funder"}],
                "investigators": [{"name": "Test Investigator"}],
                "subject_id": "invalid_subject_id",  # This will cause validation issues
            },
            "subject": {
                "subject_id": "different_subject_id",  # Mismatch will cause validation error
                "sex": "Invalid Sex Value",  # Invalid enum value
                "date_of_birth": "invalid-date-format",  # Invalid date format
            },
            "acquisition": {
                "subject_id": "yet_another_subject_id",  # Another mismatch
                "acquisition_start_time": "invalid-datetime",  # Invalid datetime
            },
        }

        from pydantic import ValidationError

        with self.assertRaises(ValidationError):
            strict_job.validate_and_create_metadata(core_metadata)

    def test_validate_and_create_metadata_success(self):
        """Test validate_and_create_metadata with valid real metadata resources"""
        data_description = self._load_resource_file(V2_METADATA_DIR, "data_description.json")
        subject = self._load_resource_file(V2_METADATA_DIR, "subject.json")
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        procedures = self._load_resource_file(V2_METADATA_DIR, "procedures.json")
        instrument = self._load_resource_file(V2_METADATA_DIR, "instrument.json")
        processing = self._load_resource_file(V2_METADATA_DIR, "processing.json")
        quality_control = self._load_resource_file(V2_METADATA_DIR, "quality_control.json")

        core_metadata = {
            "data_description": data_description,
            "subject": subject,
            "acquisition": acquisition,
            "procedures": procedures,
            "instrument": instrument,
            "processing": processing,
            "quality_control": quality_control,
        }

        metadata = self.job.validate_and_create_metadata(core_metadata)

        self.assertIsInstance(metadata, Metadata)

    def test_validate_and_create_metadata_failure_fallback(self):
        """Test validate_and_create_metadata fallback using real metadata with invalid field"""
        data_description = self._load_resource_file(V2_METADATA_DIR, "data_description.json")
        subject = self._load_resource_file(V2_METADATA_DIR, "subject.json")
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        procedures = self._load_resource_file(V2_METADATA_DIR, "procedures.json")
        instrument = self._load_resource_file(V2_METADATA_DIR, "instrument.json")
        processing = self._load_resource_file(V2_METADATA_DIR, "processing.json")
        quality_control = self._load_resource_file(V2_METADATA_DIR, "quality_control.json")

        del data_description["subject_id"]

        core_metadata = {
            "data_description": data_description,
            "subject": subject,
            "acquisition": acquisition,
            "procedures": procedures,
            "instrument": instrument,
            "processing": processing,
            "quality_control": quality_control,
        }

        metadata = self.job.validate_and_create_metadata(core_metadata)

        self.assertIsInstance(metadata, dict)

    def test_validate_and_get_subject_id_from_real_acquisition(self):
        """Test _validate_and_get_subject_id with real acquisition metadata"""
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")

        # Test that it extracts subject_id correctly
        subject_id = self.job._validate_and_get_subject_id(acquisition)
        self.assertEqual(subject_id, "804670")

    def test_validate_acquisition_start_time_with_real_data(self):
        """Test _validate_acquisition_start_time with real acquisition metadata"""
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        acq_start_time_str = acquisition["acquisition_start_time"]

        # Test that it validates correctly
        result = self.job._validate_acquisition_start_time(acq_start_time_str)
        self.assertIsInstance(result, str)
        self.assertIn("2025-09-17", result)

    def test_validate_and_get_subject_id_mismatch_with_real_data(self):
        """Test _validate_and_get_subject_id raises error with mismatched subject_id"""
        acquisition = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        acq_start_time_str = acquisition["acquisition_start_time"].replace("Z", "+00:00")
        acq_start_time = datetime.fromisoformat(acq_start_time_str)

        with patch("os.makedirs"):
            test_settings = JobSettings(
                metadata_dir="/test/metadata",
                output_dir="/test/output",
                subject_id="123456",  # Different from acquisition (804670)
                data_description_settings=DataDescriptionSettings(
                    project_name="Visual Behavior",
                    modalities=[Modality.BEHAVIOR, Modality.ECEPHYS],
                ),
                metadata_service_url="http://test-service.com",
                acquisition_start_time=acq_start_time,
                raise_if_invalid=True,
            )
            test_job = GatherMetadataJob(settings=test_settings)

        # Should raise ValueError due to subject_id mismatch
        with self.assertRaises(ValueError) as context:
            test_job._validate_and_get_subject_id(acquisition)

        self.assertIn("subject_id from acquisition metadata", str(context.exception))
        self.assertIn("does not match", str(context.exception))

    def test_run_job_pulls_subject_id_and_acquisition_start_time_from_file(self):
        """Test that subject_id and acquisition_start_time are pulled from acquisition.json when None in settings"""
        with patch("os.makedirs"):
            test_settings = JobSettings(
                metadata_dir="/test/metadata",
                output_dir="/test/output",
                subject_id=None,
                acquisition_start_time=None,
                data_description_settings=DataDescriptionSettings(
                    project_name="Visual Behavior",
                    modalities=[Modality.BEHAVIOR],
                ),
            )
            test_job = GatherMetadataJob(settings=test_settings)

        acquisition_data = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")

        with patch.object(test_job, "get_acquisition", return_value=acquisition_data):
            with patch.object(test_job, "build_data_description", return_value={"name": "test"}) as mock_build:
                with patch.object(test_job, "add_core_metadata", return_value={}) as mock_add:
                    with patch.object(test_job, "validate_and_create_metadata"):
                        with patch.object(test_job, "_write_json_file"):
                            test_job.run_job()

        mock_build.assert_called_once_with(
            acquisition_start_time=acquisition_data["acquisition_start_time"], subject_id="804670"
        )
        mock_add.assert_called_once()
        call_args = mock_add.call_args
        self.assertEqual(call_args.kwargs["subject_id"], "804670")

    def test_run_job_uses_acquisition_start_time_from_settings_when_not_in_acquisition(self):
        """Test that acquisition_start_time from settings is used when not in acquisition.json"""
        acq_start_time = datetime(2025, 9, 17, 10, 26, 0, tzinfo=timezone.utc)
        with patch("os.makedirs"):
            test_settings = JobSettings(
                metadata_dir="/test/metadata",
                output_dir="/test/output",
                subject_id="804670",
                acquisition_start_time=acq_start_time,
                data_description_settings=DataDescriptionSettings(
                    project_name="Visual Behavior",
                    modalities=[Modality.BEHAVIOR],
                ),
            )
            test_job = GatherMetadataJob(settings=test_settings)

        acquisition_data = self._load_resource_file(V2_METADATA_DIR, "acquisition.json")
        acquisition_without_time = acquisition_data.copy()
        del acquisition_without_time["acquisition_start_time"]

        with patch.object(test_job, "get_acquisition", return_value=acquisition_without_time):
            with patch.object(test_job, "build_data_description", return_value={"name": "test"}) as mock_build:
                with patch.object(test_job, "add_core_metadata", return_value={}):
                    with patch.object(test_job, "validate_and_create_metadata"):
                        with patch.object(test_job, "_write_json_file"):
                            with patch("logging.info") as mock_log_info:
                                test_job.run_job()

        mock_build.assert_called_once_with(acquisition_start_time=acq_start_time.isoformat(), subject_id="804670")
        log_calls = [call[0][0] for call in mock_log_info.call_args_list]
        self.assertTrue(any("No acquisition_start_time found in acquisition metadata" in call for call in log_calls))

    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob." + "_does_file_exist_in_user_defined_dir")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob." + "_get_file_from_user_defined_directory")
    def test_add_core_metadata_with_model(self, mock_get_file, mock_file_exists):
        """Test that add_core_metadata correctly retrieves and writes model metadata"""
        model_data = self._load_resource_file(V2_METADATA_DIR, "model.json")

        def mock_exists(file_name):
            """Mock file existence check"""
            return file_name == "model.json"

        mock_file_exists.side_effect = mock_exists

        def mock_read_file(file_name):
            """Mock file reading"""
            if file_name == "model.json":
                return model_data
            raise FileNotFoundError(f"File {file_name} not found")

        mock_get_file.side_effect = mock_read_file

        core_metadata = {}

        with patch.object(self.job, "_write_json_file") as mock_write:
            result = self.job.add_core_metadata(core_metadata=core_metadata, subject_id="804670")

        self.assertIn("model", result)
        self.assertEqual(result["model"], model_data)
        mock_write.assert_any_call("model.json", model_data)

    @patch("aind_metadata_mapper.gather_metadata.get_instrument")
    def test_get_instrument_from_service(self, mock_get_instrument):
        """Test get_instrument_from_service retrieves and writes instrument metadata"""
        instrument_data = self._load_resource_file(METADATA_SERVICE_DIR, "instrument_response.json")
        from aind_data_schema.core.instrument import Instrument

        mock_get_instrument.return_value = Instrument.model_validate(instrument_data)

        with patch("os.makedirs"):
            test_job = GatherMetadataJob(
                settings=JobSettings(
                    metadata_dir="/test/metadata",
                    output_dir="/test/output",
                    subject_id="804670",
                    data_description_settings=DataDescriptionSettings(
                        project_name="Visual Behavior",
                        modalities=[Modality.FIB],
                    ),
                    instrument_settings=InstrumentSettings(instrument_id="13A"),
                )
            )

        with patch.object(test_job, "_write_json_file") as mock_write:
            test_job.get_instrument_from_service()

        mock_get_instrument.assert_called_once()
        mock_write.assert_called_once()
        call_kwargs = mock_write.call_args[1]
        self.assertEqual(call_kwargs["filename"], "instrument_fib.json")
        self.assertIn("instrument_id", call_kwargs["contents"])
        self.assertEqual(call_kwargs["output_dir"], False)

    @patch("builtins.open", new_callable=unittest.mock.mock_open)
    @patch("os.makedirs")
    def test_write_json_file_to_metadata_dir(self, mock_makedirs, mock_open):
        """Test _write_json_file writes to metadata_dir when output_dir=False"""
        test_job = GatherMetadataJob(
            settings=JobSettings(
                metadata_dir="/test/metadata",
                output_dir="/test/output",
                subject_id="804670",
                data_description_settings=DataDescriptionSettings(project_name="Test", modalities=[Modality.FIB]),
            )
        )
        test_job._write_json_file("test.json", {"key": "value"}, output_dir=False)
        mock_open.assert_called_once_with("/test/metadata/test.json", "w")

    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob.get_acquisition")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob.build_data_description")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob.add_core_metadata")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob.validate_and_create_metadata")
    @patch("aind_metadata_mapper.gather_metadata.GatherMetadataJob._write_json_file")
    @patch("os.makedirs")
    def test_run_job_sets_metadata_dir_from_output_dir(
        self, mock_makedirs, mock_write, mock_validate, mock_add, mock_build, mock_get_acq
    ):
        """Test run_job sets metadata_dir from output_dir when None"""
        mock_get_acq.return_value = {"subject_id": "804670", "acquisition_start_time": "2025-09-17T10:26:00+00:00"}
        mock_build.return_value = {"name": "test"}
        mock_add.return_value = {}

        test_job = GatherMetadataJob(
            settings=JobSettings(
                metadata_dir=None,
                output_dir="/test/output",
                subject_id="804670",
                data_description_settings=DataDescriptionSettings(project_name="Test", modalities=[Modality.FIB]),
            )
        )
        test_job.run_job()
        self.assertEqual(test_job.settings.metadata_dir, "/test/output")


if __name__ == "__main__":
    unittest.main()



"""Script to interactively create an FIP ophys instrument.json file.

This script prompts for rig-specific information (rig ID, computer name, serial numbers)
and creates a complete instrument.json file. Serial numbers and computer names are
automatically loaded from previous instruments for the same rig ID if available.

Usage:
    First-time setup (create conda environment and install dependencies):
        conda create -n fip-instrument python=3.10 -y
        conda activate fip-instrument
        cd /path/to/aind-metadata-mapper
        pip install -e .

    Run the script:
        conda activate fip-instrument
        cd /path/to/aind-metadata-mapper
        python -m aind_metadata_mapper.fip.make_instrument

    The script will:
        1. Prompt for an Instrument ID (required, no default)
        2. Load previous instrument data for that ID if available from database
        3. Prompt for computer name (defaults to system hostname or previous value)
        4. Prompt for serial numbers (defaults to previous values if available)
        5. POST the instrument to the database and verify with a round-trip test
"""

import logging
import socket
import sys
from datetime import date
from typing import Optional

import aind_data_schema.components.devices as devices
import aind_data_schema.core.instrument as instrument
from aind_data_schema.components.connections import Connection
from aind_data_schema.components.devices import Computer
from aind_data_schema_models.modalities import Modality

from aind_metadata_mapper.utils import (
    check_existing_instrument,
    check_instrument_id,
    prompt_for_string,
    save_instrument,
)

logger = logging.getLogger(__name__)


def create_instrument(
    instrument_id: str,
    location: Optional[str] = None,
    computer_name: Optional[str] = None,
    detector_1_serial: Optional[str] = None,
    detector_2_serial: Optional[str] = None,
    objective_serial: Optional[str] = None,
    previous_instrument: Optional[instrument.Instrument] = None,
    input_func=input,
) -> instrument.Instrument:
    """Create an FIP instrument interactively.

    Parameters
    ----------
    instrument_id : str
        Instrument identifier.
    location : Optional[str]
        Location. If None, will prompt (optional).
    computer_name : Optional[str]
        Computer name. If None, defaults to system hostname or prompts.
    detector_1_serial : Optional[str]
        Green CMOS serial number. If None, will prompt (required if no previous instrument).
    detector_2_serial : Optional[str]
        Red CMOS serial number. If None, will prompt (required if no previous instrument).
    objective_serial : Optional[str]
        Objective serial number. If None, will prompt (required if no previous instrument).
    previous_instrument : Optional[Instrument]
        Optional previous instrument. Used for defaults.

    Returns
    -------
    instrument.Instrument
        Created instrument object.

    Notes
    -----
    Prompted values (if parameter is None):
        - location (optional, can be empty)
        - computer_name (defaults to system hostname if not provided)
        - detector_1_serial (Green CMOS serial number, required if no previous instrument)
        - detector_2_serial (Red CMOS serial number, required if no previous instrument)
        - objective_serial (required if no previous instrument)

    Hard-coded values (must be updated in this file if changed):
        - All device specifications (patch cords, LEDs, detectors, objective, filters, lens, harp devices)
        - Detector settings (bin_width=4, bin_height=4, crop offsets, gain=0, etc.)
        - Connection details (source_port="COM14", etc.)
        - Coordinate system (bregma, matching behavior)
        - Modality (FIB)
        - modification_date (date.today())
    """
    # Get previous instrument components for defaults
    components_by_name = {}
    if previous_instrument:
        components_by_name = {
            getattr(component, "name", ""): component
            for component in previous_instrument.components
            if getattr(component, "name", None)
        }

    # Get defaults from previous instrument or system
    defaults = {
        "location": previous_instrument.location if previous_instrument else None,
        "computer_name": socket.gethostname(),
        "detector_1_serial": getattr(components_by_name.get("Green CMOS"), "serial_number", None),
        "detector_2_serial": getattr(components_by_name.get("Red CMOS"), "serial_number", None),
        "objective_serial": getattr(components_by_name.get("Objective"), "serial_number", None),
    }

    # Use provided values or prompt for missing ones
    if location is None:
        location = prompt_for_string("Location", defaults["location"], required=False, input_func=input_func)
    if computer_name is None:
        computer_name = prompt_for_string(
            "Computer name", defaults["computer_name"], required=False, input_func=input_func
        )
    if detector_1_serial is None:
        required = defaults["detector_1_serial"] is None
        detector_1_serial = prompt_for_string(
            "Green CMOS serial number", defaults["detector_1_serial"], required=required, input_func=input_func
        )
    if detector_2_serial is None:
        required = defaults["detector_2_serial"] is None
        detector_2_serial = prompt_for_string(
            "Red CMOS serial number", defaults["detector_2_serial"], required=required, input_func=input_func
        )
    if objective_serial is None:
        required = defaults["objective_serial"] is None
        objective_serial = prompt_for_string(
            "Objective serial number", defaults["objective_serial"], required=required, input_func=input_func
        )

    # Use defaults for empty strings
    if not computer_name:
        computer_name = socket.gethostname()
    location = location if location else None

    computer = Computer(name=computer_name)

    patch_cord_note = (
        "All four patch cords are actually a single device at the camera end "
        "with four separate connections to connect to up to four implanted fibers. "
        "Unused patch cables are not physically connected to an implanted fiber during an experiment."
    )

    patch_cord_0 = devices.FiberPatchCord(
        name="Patch Cord 0",
        manufacturer=devices.Organization.DORIC,
        model="BBP(4)_200/220/900-0.37_Custom_FCM-4xMF1.25",
        core_diameter=200,
        numerical_aperture=0.37,
        notes=patch_cord_note,
    )

    patch_cord_1 = devices.FiberPatchCord(
        name="Patch Cord 1",
        manufacturer=devices.Organization.DORIC,
        model="BBP(4)_200/220/900-0.37_Custom_FCM-4xMF1.25",
        core_diameter=200,
        numerical_aperture=0.37,
        notes=patch_cord_note,
    )

    patch_cord_2 = devices.FiberPatchCord(
        name="Patch Cord 2",
        manufacturer=devices.Organization.DORIC,
        model="BBP(4)_200/220/900-0.37_Custom_FCM-4xMF1.25",
        core_diameter=200,
        numerical_aperture=0.37,
        notes=patch_cord_note,
    )

    patch_cord_3 = devices.FiberPatchCord(
        name="Patch Cord 3",
        manufacturer=devices.Organization.DORIC,
        model="BBP(4)_200/220/900-0.37_Custom_FCM-4xMF1.25",
        core_diameter=200,
        numerical_aperture=0.37,
        notes=patch_cord_note,
    )

    light_source_1 = devices.LightEmittingDiode(
        name="470nm LED",
        manufacturer=devices.Organization.THORLABS,
        model="M470F3",
        wavelength=470,
    )

    light_source_2 = devices.LightEmittingDiode(
        name="415nm LED",
        manufacturer=devices.Organization.THORLABS,
        model="M415F3",
        wavelength=415,
    )

    light_source_3 = devices.LightEmittingDiode(
        name="565nm LED",
        manufacturer=devices.Organization.THORLABS,
        model="M565F3",
        wavelength=565,
    )

    detector_1 = devices.Detector(
        name="Green CMOS",
        serial_number=detector_1_serial,
        manufacturer=devices.Organization.FLIR,
        model="BFS-U3-20S40M",
        detector_type="Camera",
        data_interface="USB",
        cooling="Air",
        immersion="air",
        bin_width=4,
        bin_height=4,
        bin_mode="Additive",
        crop_offset_x=104,
        crop_offset_y=56,
        crop_width=200,
        crop_height=200,
        gain=0,
        chroma="Monochrome",
        bit_depth=16,
    )

    detector_2 = devices.Detector(
        name="Red CMOS",
        serial_number=detector_2_serial,
        manufacturer=devices.Organization.FLIR,
        model="BFS-U3-20S40M",
        detector_type="Camera",
        data_interface="USB",
        cooling="Air",
        immersion="air",
        bin_width=4,
        bin_height=4,
        bin_mode="Additive",
        crop_offset_x=76,
        crop_offset_y=56,
        crop_width=200,
        crop_height=200,
        gain=0,
        chroma="Monochrome",
        bit_depth=16,
    )

    objective = devices.Objective(
        name="Objective",
        serial_number=objective_serial,
        manufacturer=devices.Organization.NIKON,
        model="CFI Plan Apochromat Lambda D 10x",
        numerical_aperture=0.45,
        magnification=10,
        immersion="air",
    )

    filter_1 = devices.Filter(
        name="Green emission filter",
        manufacturer=devices.Organization.SEMROCK,
        model="FF01-520/35-25",
        filter_type="Band pass",
        center_wavelength=520,
    )

    filter_2 = devices.Filter(
        name="Red emission filter",
        manufacturer=devices.Organization.SEMROCK,
        model="FF01-600/37-25",
        filter_type="Band pass",
        center_wavelength=600,
    )

    filter_3 = devices.Filter(
        name="Emission Dichroic",
        model="FF562-Di03-25x36",
        manufacturer=devices.Organization.SEMROCK,
        filter_type="Dichroic",
        cut_off_wavelength=562,
    )

    filter_4 = devices.Filter(
        name="dual-edge standard epi-fluorescence dichroic beamsplitter",
        model="FF493/574-Di01-25x36",
        manufacturer=devices.Organization.SEMROCK,
        notes="BrightLine dual-edge standard epi-fluorescence dichroic beamsplitter",
        filter_type="Multiband",
        center_wavelength=[493, 574],
    )

    filter_5 = devices.Filter(
        name="Excitation filter 410nm",
        manufacturer=devices.Organization.THORLABS,
        model="FB410-10",
        filter_type="Band pass",
        center_wavelength=410,
    )

    filter_6 = devices.Filter(
        name="Excitation filter 470nm",
        manufacturer=devices.Organization.THORLABS,
        model="FB470-10",
        filter_type="Band pass",
        center_wavelength=470,
    )

    filter_7 = devices.Filter(
        name="Excitation filter 560nm",
        manufacturer=devices.Organization.THORLABS,
        model="FB560-10",
        filter_type="Band pass",
        center_wavelength=560,
    )

    filter_8 = devices.Filter(
        name="450 Dichroic Longpass Filter",
        manufacturer=devices.Organization.EDMUND_OPTICS,
        model="#69-898",
        filter_type="Dichroic",
        cut_off_wavelength=450,
    )

    filter_9 = devices.Filter(
        name="500 Dichroic Longpass Filter",
        manufacturer=devices.Organization.EDMUND_OPTICS,
        model="#69-899",
        filter_type="Dichroic",
        cut_off_wavelength=500,
    )

    lens = devices.Lens(
        manufacturer=devices.Organization.THORLABS,
        model="AC254-080-A-ML",
        name="Image focusing lens",
    )

    cuttlefish = devices.HarpDevice(
        name="cuTTLefishFip",
        harp_device_type=devices.HarpDeviceType.CUTTLEFISHFIP,
        is_clock_generator=False,
        data_interface=devices.DataInterface.USB,
    )

    white_rabbit = devices.HarpDevice(
        name="harp_clock_generator",
        harp_device_type=devices.HarpDeviceType.WHITERABBIT,
        manufacturer=devices.Organization.AIND,
        is_clock_generator=True,
        channels=[
            devices.DAQChannel(channel_name="ClkOut", channel_type=devices.DaqChannelType.DO),
        ],
    )

    connections = [
        Connection(
            source_device="cuTTLefishFip",
            source_port="COM14",
            target_device=computer_name,
        ),
        Connection(
            source_device="harp_clock_generator",
            source_port="ClkOut",
            target_device="cuTTLefishFip",
            target_port="COM14",
            send_and_receive=False,
        ),
    ]

    # Coordinate system matching behavior (bregma with X/Y/Z axes, not BREGMA_ARI)
    coordinate_system = {
        "object_type": "Coordinate system",
        "name": "origin",
        "origin": "Bregma",
        "axes": [
            {"object_type": "Axis", "name": "X", "direction": "Left_to_right"},
            {"object_type": "Axis", "name": "Y", "direction": "Anterior_to_posterior"},
            {"object_type": "Axis", "name": "Z", "direction": "Inferior_to_superior"},
        ],
        "axis_unit": "millimeter",
    }

    instrument_model = instrument.Instrument(
        location=location if location else None,
        instrument_id=instrument_id,
        modification_date=date.today(),
        modalities=[Modality.FIB],
        coordinate_system=coordinate_system,
        components=[
            computer,
            patch_cord_0,
            patch_cord_1,
            patch_cord_2,
            patch_cord_3,
            light_source_1,
            light_source_2,
            light_source_3,
            detector_1,
            detector_2,
            objective,
            filter_1,
            filter_2,
            filter_3,
            filter_4,
            filter_5,
            filter_6,
            filter_7,
            filter_8,
            filter_9,
            lens,
            cuttlefish,
            white_rabbit,
        ],
        calibrations=[],
        connections=connections,
    )

    return instrument_model


def main(
    instrument_id: str,
    skip_confirmation: bool = False,
    input_func=input,
) -> None:
    """Main function to create an instrument interactively.

    Parameters
    ----------
    instrument_id : str
        Instrument ID (required).
    skip_confirmation : bool
        If True, skip confirmation prompt for new instrument IDs.
    """
    # Check if instrument exists and get previous instrument data
    previous_instrument = check_instrument_id(
        instrument_id=instrument_id,
        skip_confirmation=skip_confirmation,
        input_func=input_func,
    )

    # Create instrument interactively
    instrument_model = create_instrument(
        instrument_id,
        previous_instrument=previous_instrument,
        input_func=input_func,
    )

    # Check if instrument with same ID and date already exists
    modification_date_str = instrument_model.modification_date.isoformat()
    record_exists = check_existing_instrument(instrument_id, modification_date_str)
    replace = False
    if record_exists:
        logger.info(
            f"An instrument with ID '{instrument_id}' and modification_date '{modification_date_str}' already exists."
        )
        response = input_func("Do you want to overwrite the existing record? [y/N]: ").strip().lower()
        if response in ("y", "yes"):
            replace = True
            logger.info("Will overwrite existing record.")
        else:
            logger.info("Cancelled. Not overwriting existing record.")
            sys.exit(0)

    # Save instrument (includes round-trip validation)
    try:
        save_instrument(instrument_model, replace=replace)
    except ValueError as e:
        logger.error(f"Failed - {e}")
        sys.exit(1)


if __name__ == "__main__":
    instrument_id = prompt_for_string("Instrument ID", required=True)
    main(instrument_id=instrument_id)



"""Utility functions for AIND metadata mappers."""

import json
import logging
import sys
import time
from datetime import date, datetime
from pathlib import Path
from typing import Optional
from urllib.parse import urljoin

import aind_data_schema.core.instrument as instrument
import requests
import yaml

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

INSTRUMENT_BASE_URL = "http://aind-metadata-service/api/v2/instrument"
PROCEDURES_BASE_URL = "http://aind-metadata-service/api/v2/procedures"
SUBJECT_BASE_URL = "http://aind-metadata-service/api/v2/subject"


def normalize_utc_timezone(dt: str) -> str:
    """
    Normalize UTC timezone indicator from 'Z' to '+00:00' offset format.

    This ensures compatibility with Python 3.10's datetime.fromisoformat(),
    which doesn't support the 'Z' shorthand for UTC.

    Parameters
    ----------
    dt : str
        An ISO 8601 datetime string, potentially ending with 'Z'.

    Returns
    -------
    str
        The datetime string with 'Z' replaced by '+00:00' if present,
        otherwise unchanged.

    Examples
    --------
    >>> normalize_utc_timezone("2025-11-16T23:00:22Z")
    '2025-11-16T23:00:22+00:00'
    >>> normalize_utc_timezone("2025-11-16T23:00:22-05:00")
    '2025-11-16T23:00:22-05:00'
    """
    if dt.endswith("Z"):
        return dt[:-1] + "+00:00"
    return dt


def ensure_timezone(dt):
    """Ensure datetime has timezone info using system local timezone.

    Parameters
    ----------
    dt : datetime, str, or None
        Datetime to process. Can be a datetime object, ISO format string, or None.

    Returns
    -------
    datetime
        Datetime with timezone info. If None is provided, returns current time
        in local timezone.
    """
    if dt is None:
        return datetime.now().astimezone()
    if isinstance(dt, str):
        dt = datetime.fromisoformat(dt)
    if dt.tzinfo is None:
        # Use system's local timezone
        local_tz = datetime.now().astimezone().tzinfo
        dt = dt.replace(tzinfo=local_tz)
    return dt


def metadata_service_helper(
    url: str, timeout: int = 60, max_retries: int = 3, retry_delay: float = 1.0
) -> Optional[dict]:
    """Fetch metadata from a service URL with retry logic.

    Parameters
    ----------
    url : str
        Full URL to fetch metadata from.
    timeout : int
        Request timeout in seconds. Default is 60.
    max_retries : int
        Maximum number of retry attempts. Defaults to 3.
    retry_delay : float
        Delay in seconds between retry attempts. Defaults to 1.0.

    Returns
    -------
    Optional[dict]
        Metadata as a dictionary, or None if error occurs after all retries.
    """
    for attempt in range(max_retries):
        try:
            response = requests.get(url, timeout=timeout)
            if response.status_code == 400:
                return response.json()
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            if attempt < max_retries - 1:
                logger.warning(
                    f"Error retrieving metadata from {url} (attempt {attempt + 1}/{max_retries}): {e}. "
                    f"Retrying in {retry_delay}s..."
                )
                time.sleep(retry_delay)
            else:
                logger.error(f"Error retrieving metadata from {url} after {max_retries} attempts: {e}")
                return None

    return None


def get_subject(subject_id: str, base_url: str = SUBJECT_BASE_URL) -> Optional[dict]:
    """Fetch subject data from the metadata service.

    Parameters
    ----------
    subject_id : str
        The subject ID to query.
    base_url : str
        Base URL for the subject endpoint. Defaults to SUBJECT_BASE_URL.

    Returns
    -------
    Optional[dict]
        Subject data dictionary, or None if the request fails.
    """
    try:
        url_base = base_url.rstrip("/") + "/"
        url = urljoin(url_base, subject_id.lstrip("/"))
        result = metadata_service_helper(url)
        if result is None:
            logger.warning(f"Could not fetch subject {subject_id}")
        return result
    except Exception as e:
        logger.warning(f"Unexpected error fetching subject {subject_id}: {e}")
        return None


def get_procedures(subject_id: str, base_url: str = PROCEDURES_BASE_URL, timeout_seconds: int = 300) -> Optional[dict]:
    """Fetch procedures data for a subject from the metadata service.

    Parameters
    ----------
    subject_id : str
        The subject ID to query.
    base_url : str
        Base URL for the procedures endpoint. Defaults to PROCEDURES_BASE_URL.
    timeout_seconds : int
        Request timeout in seconds. Default is 300.
    Returns
    -------
    Optional[dict]
        Procedures data dictionary, or None if the request fails.
    """
    try:
        url_base = base_url.rstrip("/") + "/"
        url = urljoin(url_base, subject_id.lstrip("/"))
        result = metadata_service_helper(url, timeout=timeout_seconds)
        if result is None:
            logger.warning(f"Could not fetch procedures for subject {subject_id}")
        return result
    except Exception as e:
        logger.warning(f"Unexpected error fetching procedures for subject {subject_id}: {e}")
        return None


def get_intended_measurements(
    subject_id: str,
    base_url: str = "http://aind-metadata-service/intended_measurements",
) -> Optional[dict]:
    """Fetch intended measurements for a subject from the metadata service.

    Parameters
    ----------
    subject_id : str
        The subject ID to query.
    base_url : str
        Base URL for the intended measurements endpoint.

    Returns
    -------
    Optional[dict]
        Intended measurements data dictionary, or None if the request fails.
    """
    try:
        url = f"{base_url}/{subject_id}"
        result = metadata_service_helper(url, timeout=5)
        if result is None:
            logger.warning(f"Could not fetch intended measurements for subject {subject_id}")
        return result
    except Exception as e:
        logger.warning(f"Error fetching intended measurements for subject {subject_id}: {e}")
        return None


def get_protocols_for_modality(modality):
    """Get protocol URLs for a specific modality.

    Loads protocols from protocols.yaml file and returns the list for the given modality.

    Parameters
    ----------
    modality : str
        The modality name (e.g., 'fip', 'smartspim').

    Returns
    -------
    list
        List of protocol URLs for the modality.
    """
    try:
        project_root = Path(__file__).parent.parent.parent
        protocols_file = project_root / "protocols.yaml"

        if not protocols_file.exists():
            logger.warning(f"Protocols file not found at {protocols_file}")
            return []

        with open(protocols_file, "r") as f:
            protocols = yaml.safe_load(f)

        protocols = protocols or {}
        return protocols.get(modality, [])
    except Exception as e:
        logger.warning(f"Error loading protocols: {e}")
        return []


def get_instrument(
    instrument_id: str,
    modification_date: Optional[str] = None,
    suppress_warning: bool = False,
    base_url: str = INSTRUMENT_BASE_URL,
    output_directory: Optional[str | Path] = None,
    prefix: Optional[str] = None,
    filename_suffix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> Optional[instrument.Instrument]:
    """Get instrument.

    Gets the latest record by default, or a specific record if modification_date is provided.

    Parameters
    ----------
    instrument_id : str
        Instrument identifier.
    modification_date : Optional[str]
        Optional modification date (YYYY-MM-DD format). If None, returns latest record.
    suppress_warning : bool
        If True, suppress warning when modification_date not found.
    base_url : str
        Base URL for the instrument endpoint. Defaults to INSTRUMENT_BASE_URL.
    output_directory : Optional[str | Path]
        If provided, save the instrument using write_standard_file(). Passed through
        to Instrument.write_standard_file().
    prefix : Optional[str]
        Filename prefix for write_standard_file (e.g. "ephys" -> ephys_instrument.json).
    filename_suffix : Optional[str]
        Filename suffix for write_standard_file.
    suffix : Optional[str]
        File extension override for write_standard_file.

    Returns
    -------
    Optional[Instrument]
        Instrument model, or None if not found.
    """
    try:
        url_base = base_url.rstrip("/") + "/"
        url = urljoin(url_base, instrument_id.lstrip("/"))
        url_with_params = f"{url}?partial_match=true"
        result = metadata_service_helper(url_with_params)
        if result is None:
            logger.warning(f"Could not fetch instrument {instrument_id}")
            return None

        records = result if isinstance(result, list) else [result]
        matching_records = [
            r for r in records if r.get("modification_date") and r.get("instrument_id") == instrument_id
        ]
        if not matching_records:
            return None

        if modification_date:
            record = next(
                (r for r in matching_records if r.get("modification_date") == modification_date),
                None,
            )
            if record is None and not suppress_warning:
                available_dates = sorted(
                    set(r.get("modification_date") for r in matching_records if r.get("modification_date"))
                )
                logger.warning(
                    f"No record found for instrument_id '{instrument_id}' with "
                    f"modification_date '{modification_date}'. "
                    f"Available dates: {available_dates}"
                )
        else:
            record = sorted(
                matching_records,
                key=lambda r: r["modification_date"],
            )[-1]

        if record is not None and output_directory is not None:
            _write_instrument_to_path(
                record,
                output_directory=output_directory,
                prefix=prefix,
                filename_suffix=filename_suffix,
                suffix=suffix,
            )
        if record is not None:
            return instrument.Instrument.model_validate(record)
        return None
    except Exception as e:
        logger.warning(f"Unexpected error fetching instrument {instrument_id}: {e}")
        return None


def _write_instrument_to_path(
    record: dict,
    output_directory: str | Path,
    prefix: Optional[str] = None,
    filename_suffix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> None:
    """Save instrument record using write_standard_file."""
    inst = instrument.Instrument.model_validate(record)
    path = Path(output_directory)
    path.mkdir(parents=True, exist_ok=True)
    inst.write_standard_file(
        output_directory=path,
        prefix=prefix,
        filename_suffix=filename_suffix,
        suffix=suffix,
    )


def save_instrument(
    instrument_model: instrument.Instrument | dict | str | Path,
    replace: bool = False,
    update_modification_date: bool = True,
) -> None:
    """Save instrument and validate round-trip.

    Saves the instrument, then retrieves it back and verifies that what we get back
    matches what we just saved. If so, we know the save worked.

    Parameters
    ----------
    instrument_model : instrument.Instrument | dict | str | Path
        Instrument to POST. Can be an Instrument object, a dictionary, or a path to
        a JSON file. If a path (str or Path) is provided, the file is loaded and validated.
    replace : bool
        If True, overwrite existing record with same instrument_id and modification_date.
    update_modification_date : bool
        If True (default), set modification_date to today (YYYY-MM-DD). If False,
        keep the modification date as passed in the instrument.

    Raises
    ------
    ValueError
        If POST fails (e.g., record already exists) or round-trip validation fails.
    requests.HTTPError
        If server error occurs (500+).
    """
    # Load from file if path provided
    if isinstance(instrument_model, (str, Path)):
        with open(instrument_model) as f:
            instrument_model = json.load(f)
    # Convert dict to Instrument object if needed (also validates the data)
    if isinstance(instrument_model, dict):
        instrument_model = instrument.Instrument.model_validate(instrument_model)

    if update_modification_date:
        instrument_model = instrument_model.model_copy(update={"modification_date": date.today()})

    # Use model_dump_json() and parse to ensure dates are properly serialized
    source_dict = json.loads(instrument_model.model_dump_json())
    logger.info(f"POSTing instrument to {INSTRUMENT_BASE_URL}")
    params = {"replace": "true"} if replace else {}
    response = requests.post(INSTRUMENT_BASE_URL, json=source_dict, params=params)
    # POST 400 is always an error (e.g., "Record already exists")
    if response.status_code == 400:
        error_msg = response.json().get("message", response.text)
        raise ValueError(f"Cannot POST instrument: {error_msg}")
    if response.status_code >= 500:
        response.raise_for_status()

    # GET back and validate round-trip
    logger.info(
        f"GETting instrument from {INSTRUMENT_BASE_URL}/{instrument_model.instrument_id} "
        "to verify that save was successful"
    )
    read_back_instrument = get_instrument(instrument_model.instrument_id)
    if read_back_instrument is None:
        raise ValueError(f"Instrument '{instrument_model.instrument_id}' not found in database")

    # Validate round-trip
    read_back_dict = json.loads(read_back_instrument.model_dump_json())
    if source_dict == read_back_dict:
        logger.info("Instrument.json successfully stored in the db")
    else:
        raise ValueError("Round-trip test failed: Source and read-back instruments differ")


def check_existing_instrument(
    instrument_id: str,
    modification_date: str,
) -> bool:
    """Check if an instrument with the same ID and modification_date already exists.

    Parameters
    ----------
    instrument_id : str
        Instrument identifier.
    modification_date : str
        Modification date (YYYY-MM-DD format).

    Returns
    -------
    bool
        True if a record with the same instrument_id and modification_date exists.
    """
    existing = get_instrument(
        instrument_id,
        modification_date=modification_date,
        suppress_warning=True,
    )
    return existing is not None


def check_instrument_id(
    instrument_id: str,
    skip_confirmation: bool = False,
    input_func=input,
) -> Optional[instrument.Instrument]:
    """Check if instrument exists and get previous instrument data.

    Checks if records exist for the given instrument_id and returns the
    previous instrument if found. Confirms creation if instrument doesn't exist.

    Parameters
    ----------
    instrument_id : str
        Instrument ID to check.
    skip_confirmation : bool
        If True, skip confirmation prompt for new instrument IDs.
    input_func : callable
        Function to get user input. Defaults to builtin input().

    Returns
    -------
    Optional[Instrument]
        Previous instrument, or None if not found.

    Raises
    ------
    SystemExit
        If user cancels creation of new instrument ID.
    """
    # Check if instrument exists
    previous_instrument = get_instrument(instrument_id)
    if previous_instrument is None and not skip_confirmation:
        logger.info("This is a new instrument ID.")
        response = (
            input_func(f"Are you sure you want to create a new ID with name '{instrument_id}'? [Y/n]: ").strip().lower()
        )
        if response and response not in ("y", "yes"):
            logger.info("Cancelled. Please run the script again with the correct instrument ID.")
            sys.exit(0)

    return previous_instrument


def prompt_for_string(
    prompt: str,
    default: Optional[str] = None,
    required: bool = False,
    help_message: Optional[str] = None,
    input_func=input,
) -> str:
    """Prompt user for a string value.

    Parameters
    ----------
    prompt : str
        The prompt message.
    default : Optional[str]
        Optional default value to display.
    required : bool
        If True and no default provided, require non-empty input.
    help_message : Optional[str]
        Optional help message to display when required field is empty.
    input_func : callable
        Function to get user input. Defaults to builtin input().

    Returns
    -------
    str
        User input or default if provided and user presses enter.
    """
    while True:
        if default:
            full_prompt = f"{prompt} [{default}]: "
        else:
            full_prompt = f"{prompt}: "
        response = input_func(full_prompt).strip()
        if response:
            return response
        if default:
            return default
        if not required:
            return ""
        print("This field is required. Please enter a value.")
        if help_message:
            print(help_message)



"""Individual mappers should be registered here by name

Each JobClass must inherit from MapperJob in base.py and accept a JobSettings subclass
"""

from aind_metadata_mapper.fip.mapper import FIPMapper

registry = {
    "fip": FIPMapper,
}



"""Tests class in gather_processing_job module."""

import unittest
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, call, mock_open, patch

from aind_data_schema.base import GenericModel
from aind_data_schema.components.identifiers import Code
from aind_data_schema.core.processing import (
    DataProcess,
    Processing,
    ProcessStage,
)
from aind_data_schema_models.process_names import ProcessName

from aind_metadata_mapper.gather_processing_job import (
    GatherProcessingJob,
    JobSettings,
)


class TestGatherProcessingJob(unittest.TestCase):
    """Tests methods in GatherProcessingJob class."""

    @patch("pathlib.Path.exists")
    @patch("builtins.open", new_callable=mock_open, read_data='{"data_processes": []}')
    @patch("json.load")
    def test_load_existing_processing_exists_and_valid(
        self, mock_json_load: MagicMock, mock_file_open: MagicMock, mock_exists: MagicMock
    ):
        """Tests load_existing_processing when file exists and is valid."""
        mock_exists.return_value = True
        existing_processing_dict = {
            "data_processes": [
                {
                    "start_date_time": "2024-09-09T01:02:03",
                    "end_date_time": "2024-09-10T01:02:03",
                    "process_type": "Image atlas alignment",
                    "experimenters": ["Existing Experimenter"],
                    "stage": "Processing",
                    "code": {"url": "www.example.com/existing"},
                }
            ]
        }
        mock_json_load.return_value = existing_processing_dict

        example_processing = Processing(data_processes=[])
        example_settings = JobSettings(output_directory="example", processing=example_processing)
        job = GatherProcessingJob(settings=example_settings)

        result = job.load_existing_processing()

        self.assertIsNotNone(result)
        self.assertIsInstance(result, Processing)
        self.assertEqual(len(result.data_processes), 1)
        mock_exists.assert_called_once()
        mock_json_load.assert_called_once()

    @patch("pathlib.Path.exists")
    def test_load_existing_processing_does_not_exist(self, mock_exists: MagicMock):
        """Tests load_existing_processing when file does not exist."""
        mock_exists.return_value = False

        example_processing = Processing(data_processes=[])
        example_settings = JobSettings(output_directory="example", processing=example_processing)
        job = GatherProcessingJob(settings=example_settings)

        result = job.load_existing_processing()

        self.assertIsNone(result)
        mock_exists.assert_called_once()

    @patch("pathlib.Path.exists")
    @patch("builtins.open", new_callable=mock_open)
    @patch("json.load")
    @patch("logging.error")
    def test_load_existing_processing_invalid_schema(
        self, mock_log_error: MagicMock, mock_json_load: MagicMock, mock_file_open: MagicMock, mock_exists: MagicMock
    ):
        """Tests load_existing_processing when file exists but doesn't match Processing schema."""
        mock_exists.return_value = True
        mock_json_load.return_value = {"invalid_field": "value"}

        example_processing = Processing(data_processes=[])
        example_settings = JobSettings(output_directory="example", processing=example_processing)
        job = GatherProcessingJob(settings=example_settings)

        with self.assertRaises(Exception):
            job.load_existing_processing()

        mock_log_error.assert_called_once()

    @patch("builtins.open", new_callable=mock_open)
    @patch("json.dump")
    def test_run_job(self, mock_json_dump: MagicMock, mock_file_open: MagicMock):
        """Tests run_job method with example processing data."""
        # noinspection PyArgumentList
        example_processing = Processing(
            data_processes=[
                DataProcess(
                    start_date_time=datetime(2024, 10, 10, 1, 2, 3),
                    end_date_time=datetime(2024, 10, 11, 1, 2, 3),
                    process_type=ProcessName.COMPRESSION,
                    experimenters=["AIND Scientific Computing"],
                    stage=ProcessStage.PROCESSING,
                    code=Code(
                        url="www.example.com/ephys_compression",
                        parameters=GenericModel(),
                    ),
                ),
                DataProcess(
                    start_date_time=datetime(2024, 10, 10, 1, 2, 3),
                    end_date_time=datetime(2024, 10, 11, 1, 2, 4),
                    process_type=ProcessName.OTHER,
                    experimenters=["AIND Scientific Computing"],
                    stage=ProcessStage.PROCESSING,
                    code=Code(url=""),
                    notes="Data was copied.",
                ),
            ]
        )
        example_settings = JobSettings(output_directory="example", processing=example_processing)
        job = GatherProcessingJob(settings=example_settings)
        job.run_job()
        mock_file_open.assert_has_calls(calls=[call(Path("example") / "processing.json", "w")])
        mock_json_dump.assert_called_once()

    @patch("aind_metadata_mapper.gather_processing_job.GatherProcessingJob.load_existing_processing")
    @patch("builtins.open", new_callable=mock_open)
    @patch("json.dump")
    @patch("logging.error")
    def test_run_job_merge_fails_raises_error(
        self,
        mock_log_error: MagicMock,
        mock_json_dump: MagicMock,
        mock_file_open: MagicMock,
        mock_load_existing: MagicMock,
    ):
        """Tests run_job raises exception when merge fails."""
        existing_processing = Processing(
            data_processes=[
                DataProcess(
                    start_date_time=datetime(2024, 9, 9, 1, 2, 3),
                    end_date_time=datetime(2024, 9, 10, 1, 2, 3),
                    process_type=ProcessName.IMAGE_ATLAS_ALIGNMENT,
                    experimenters=["Existing Experimenter"],
                    stage=ProcessStage.PROCESSING,
                    code=Code(url="www.example.com/existing_process"),
                )
            ]
        )
        mock_load_existing.return_value = existing_processing
        with patch.object(Processing, "__add__", side_effect=Exception("Merge failed")):
            new_processing = Processing(
                data_processes=[
                    DataProcess(
                        start_date_time=datetime(2024, 10, 10, 1, 2, 3),
                        end_date_time=datetime(2024, 10, 11, 1, 2, 3),
                        process_type=ProcessName.COMPRESSION,
                        experimenters=["AIND Scientific Computing"],
                        stage=ProcessStage.PROCESSING,
                        code=Code(url="www.example.com/ephys_compression"),
                    )
                ]
            )
            example_settings = JobSettings(output_directory="example", processing=new_processing)
            job = GatherProcessingJob(settings=example_settings)

            with self.assertRaises(Exception) as context:
                job.run_job()

            self.assertIn("Merge failed", str(context.exception))
            mock_log_error.assert_called_once()
            self.assertIn("Failed to merge existing processing.json", str(mock_log_error.call_args))
            mock_json_dump.assert_not_called()


if __name__ == "__main__":
    unittest.main()



""" "Unit tests for gather_metadata_registry.py."""

import json
import os
import tempfile
import unittest
from unittest import mock

from aind_data_schema_models.modalities import Modality

from aind_metadata_mapper.base import MapperJob, MapperJobSettings
from aind_metadata_mapper.gather_metadata import GatherMetadataJob, JobSettings
from aind_metadata_mapper.mapper_registry import registry
from aind_metadata_mapper.models import DataDescriptionSettings


class TestMapperJob(MapperJob):
    """A test mapper job that does nothing."""

    def run_job(self, job_settings: MapperJobSettings) -> None:
        """Run the test mapping job.

        This implementation does nothing.
        """
        raise NotImplementedError("This is a test mapper job and does not implement run_job.")


class TestGatherMetadataJob(unittest.TestCase):
    """Unit tests for GatherMetadataJob with registry."""

    def setUp(self):
        """Set up test environment."""
        # Create a temporary directory for metadata files
        self.temp_dir = tempfile.TemporaryDirectory()
        self.metadata_dir = self.temp_dir.name
        # Create a dummy input file for the mapper
        self.mapper_name = "test_mapper_job"
        self.input_filename = f"{self.mapper_name}.json"
        self.input_path = os.path.join(self.metadata_dir, self.input_filename)
        with open(self.input_path, "w") as f:
            json.dump({"dummy": "data"}, f)
        # Patch registry to include our test mapper
        registry[self.mapper_name] = TestMapperJob

    def tearDown(self):
        """Clean up test environment."""
        self.temp_dir.cleanup()

    @mock.patch.object(TestMapperJob, "run_job", autospec=True)
    @mock.patch("os.listdir")
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("aind_metadata_mapper.gather_metadata.logging")
    def test_run_mappers_for_acquisition_registry_key(self, mock_logging, mock_exists, mock_listdir, mock_run_job):
        """Test _run_mappers_for_acquisition with a registry key."""
        # Patch os.listdir to return our test file
        mock_listdir.return_value = [self.input_filename]

        # Provide all required JobSettings fields
        data_desc_settings = DataDescriptionSettings(
            project_name="Test Project",
            modalities=[Modality.ECEPHYS],
        )
        settings = JobSettings(
            metadata_dir=self.metadata_dir,
            output_dir=self.metadata_dir,
            subject_id="test_subject",
            data_description_settings=data_desc_settings,
        )
        job = GatherMetadataJob(settings=settings)
        job._run_mappers_for_acquisition()
        # Check that run_job was called for the registry key
        mock_run_job.assert_called_once()
        args, kwargs = mock_run_job.call_args
        # The job_settings argument should have correct input/output paths
        self.assertEqual(str(args[1].input_filepath), str(self.input_path))
        self.assertEqual(str(args[1].output_directory), str(self.metadata_dir))
        self.assertEqual(args[1].output_filename_suffix, self.mapper_name)

    @mock.patch("os.listdir")
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("aind_metadata_mapper.gather_metadata.logging")
    def test_run_mappers_for_acquisition_registry_key_raises(self, mock_logging, mock_exists, mock_listdir):
        """Test _run_mappers_for_acquisition raises NotImplementedError for test mapper."""
        # Patch os.listdir to return our test file
        mock_listdir.return_value = [self.input_filename]

        data_desc_settings = DataDescriptionSettings(
            project_name="Test Project",
            modalities=[Modality.ECEPHYS],
        )
        settings = JobSettings(
            metadata_dir=self.metadata_dir,
            output_dir=self.metadata_dir,
            subject_id="test_subject",
            data_description_settings=data_desc_settings,
            raise_if_mapper_errors=True,
        )
        job = GatherMetadataJob(settings=settings)
        with self.assertRaises(NotImplementedError):
            job._run_mappers_for_acquisition()

        settings.raise_if_mapper_errors = False
        job = GatherMetadataJob(settings=settings)
        # This should not raise now
        job._run_mappers_for_acquisition()


if __name__ == "__main__":
    registry["test_mapper_job"] = TestMapperJob
    unittest.main()



"""Tests methods in models module"""

import unittest

from aind_data_schema_models.modalities import Modality

from aind_metadata_mapper.models import DataDescriptionSettings, JobSettings


class TestJobSettings(unittest.TestCase):
    """Tests JobSettings class."""

    def test_basic_constructor_from_args(self):
        """Tests basic constructor from command line args."""
        # Create JobSettings directly with parameters since command line
        # parsing of modalities has complex requirements
        data_desc_settings = DataDescriptionSettings(
            project_name="test_project",
            modalities=[Modality.ECEPHYS, Modality.BEHAVIOR],
        )
        job_settings = JobSettings(
            metadata_dir=".",
            output_dir="./output",
            subject_id="12345",
            data_description_settings=data_desc_settings,
        )
        self.assertEqual(".", job_settings.metadata_dir)
        self.assertEqual("12345", job_settings.subject_id)
        self.assertEqual("test_project", job_settings.data_description_settings.project_name)
        expected_modalities = [Modality.ECEPHYS, Modality.BEHAVIOR]
        self.assertEqual(expected_modalities, job_settings.data_description_settings.modalities)

    def test_output_path_defaults_to_input_path(self):
        """Tests that output_dir defaults to metadata_dir when not provided."""
        data_desc_settings = DataDescriptionSettings(
            project_name="test_project",
            modalities=[Modality.ECEPHYS],
        )
        job_settings = JobSettings(
            metadata_dir="/test/input",
            output_dir="/test/input",
            subject_id="12345",
            data_description_settings=data_desc_settings,
        )
        self.assertEqual("/test/input", job_settings.metadata_dir)
        self.assertEqual("/test/input", job_settings.output_dir)


if __name__ == "__main__":
    unittest.main()
